/**
 ** David Bonnafous - Anthony Mallet - Thu Jul  5 2001
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** $Log: geom.c,v $
 ** Revision 1.5  2007/11/27 10:41:19  adegroot
 ** Clean a bit the code so it can compile with -W -Wall -Wextra -Wwrite-strings -Werror
 **
 ** -pedantic is not supported because we use gnu extension #ident
 **
 ** Revision 1.4  2007/11/22 10:06:26  adegroot
 ** Add a workaround in case where the robot is not fully in the dtm.
 **
 ** If we can't compute the configuration of back axle, create a fake configuration
 ** based on the real position of the robot.
 **
 ** Revision 1.3  2007/06/19 17:03:13  tpeynot
 ** default should not be VERBOSE
 **
 ** Revision 1.2  2007/01/08 19:55:17  sjoyeux
 ** * sync with the DEM tag (was something else before)
 ** * update to OpenRobot's T3Ds
 **
 ** Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 ** Packaged the P3D module from Lama's calife.
 ** See CHANGES for a list of modifications
 **
 **
 ** Revision 1.9.2.8  2005/07/13 15:46:49  tpeynot
 ** Add a "verbose" option for placement function
 **
 ** Revision 1.9.2.7  2005/04/26 12:30:46  tpeynot
 ** Update some printf
 ** and take into account the fact that the (z-*height)/len arg of asin may not be between -1 and 1
 **
 ** Revision 1.9.2.6  2005/04/21 12:31:29  tpeynot
 ** Add gdheDriver.c, containing functions for gdhe connection
 **
 ** Revision 1.9.2.5  2005/04/20 16:52:36  tpeynot
 ** little changes in bodyCollisionTest function
 **
 ** Revision 1.9.2.4  2005/04/13 20:27:38  tpeynot
 ** Modifications of protos and structures for body collision test (e.g. ground clearance info)
 **
 ** Revision 1.9.2.3  2005/04/07 20:42:44  tpeynot
 ** Add functions and elements in structures etc... for Dala body collision test
 ** Most of them are commented for the moment
 **
 ** Revision 1.9.2.2  2005/03/29 09:31:40  tpeynot
 ** Dala Body Collision Test now uses a DEM map
 **
 ** Revision 1.9.2.1  2005/03/18 23:42:38  tpeynot
 ** p3d uses for now an internal map: a struct P3D_MAP
 ** This struct will contain info a dem from the dem module
 **
 ** Revision 1.9  2005/03/18 09:58:14  tpeynot
 ** Just some comments
 **
 ** Revision 1.8  2005/03/17 13:46:53  tpeynot
 ** Add param "bestArc" ro p3dPlan fct to get info about the arc selectes by p3d planner
 ** Add functions adapted to Dala's structure instead of Lama. These functions are used if there's a flag -DATRV in Makefile. Otherwise, the lib will keep working with Lama adapted functions
 **
 ** Revision 1.7  2001/09/28 16:12:01  mallet
 ** Le test de validit du numro de profil de roue tait bugg...
 **
 ** Revision 1.6  2001/07/16 19:45:54  mallet
 ** fonctions de sauvegarde/restauration de mnt
 **
 ** Revision 1.5  2001/07/15 17:44:42  mallet
 ** sync
 **
 ** Revision 1.4  2001/07/14 11:39:09  mallet
 ** premiere version ok
 **
 ** Revision 1.3  2001/07/12 17:38:45  mallet
 ** sync
 **
 ** Revision 1.2  2001/07/12 09:56:00  mallet
 ** plus robuste dans la detection de la fin des iterations du placement (amortissement)
 **
 ** Revision 1.1  2001/07/11 18:05:41  mallet
 ** re-ecriture de p3d
 **
 **/
#ident "$Id: geom.c,v 1.5 2007/11/27 10:41:19 adegroot Exp $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "libp3d.h"
#include "p3dInt.h"

/*---- Useful Constants ------*/

/*-- Discretisation factor: sigma_discretisation = step / (2*sqrt(3))--*/
/*-- So this factor is an approximation of 1/(2*sqrt(3))  */
#define P3D_DISCRETISATION_FACTOR (0.29)


#define DEBUG 0

#define TEMP 0

#define VERBOSE 1  /* 1 if want a printf for unstable or unacceptable config. */

#define BODYTEST 0

static double	p3dGetConstrainedAxleRoll(P3DRobot *robot,
			P3DRobotGeometry *geom, double x, double y,
			double z, double theta, double len,
			P3D_MAP *map, double *roll, double *pitch,
			double *height, double *pUnk, double *var, double *var_z);
static double	p3dGetAxleRoll(P3DRobot *robot, P3DRobotGeometry *geom,
			double x, double y, double theta,
			P3D_MAP *map, double *roll, double *height,
			double *pUnk, double *var, double *var_z);
static double	p3dGetWheelHeight(P3DElement *wheel, double x, double y,
			double theta, P3D_MAP *map, double *wu,
			double *wv, double *wh, double *pUnk, double *sigma);
static void	p3dWarnUnstable(double x, double y, double theta, 
			const char *message);
static void init_config(P3DRobotConfig *config);

#if ((defined(ATRV) || defined(TROOPER) || defined(RMP400)) && BODYTEST)
static double
p3dBodyCollisionTest(P3DElement *body,
			 double x, double y, double z, double theta, 
			 P3D_MAP *map,
			 double *wu, double *wv, double *wh, double *pUnk);

static int
p3dCheckBodyCollision(P3DRobot *robot, P3DRobotGeometry * geom, P3D_MAP *map, 
		P3DRobotConfig *config);

#endif /* ATRV */

static void
init_config(P3DRobotConfig *config)
{
   /* Initialize config, in case it can't be computed */
   config->q = -1.0;		
   config->x = 0.0;
   config->y = 0.0;
   config->z = 0.0;
   config->theta = 0.0;
   config->pitch = 0.0;
   config->roll  = 0.0;
   config->alpha1 = 0.0;
   config->alpha2 = 0.0;
   config->beta3 = 0.0;
   config->dzBodyGround = 0.0;
   config->sigmaRoll = 0.0;
   config->sigmaPitch = 0.0;
   config->sigmaz = 0.0;
}


/*
 * p3dPlaceRobot --------------------------------------------------------
 *
 * Compute robot configuration for position "x,y", orientation "theta"
 * wrt. the center of "map".
 *
 * Return the quality of placement (minimum of the one of the three
 * axles). "roll" and "pitch" are the configuration of the middle axle,
 * "alpha1" and "alpha2" the *relative* roll of the front and back axle
 * and "beta" the angle between the three axles. pUnk is the max
 * percentage of unknown terrain for each wheel.
 */

double
p3dPlaceRobot(P3DRobot *robot, P3DRobotGeometry *geom,
	      double x, double y, double theta, P3D_MAP *map,
	      P3DRobotConfig *config)
{
   double fQ, mQ, bQ;
   double fpUnk, mpUnk, bpUnk;
   double fRoll, mRoll, bRoll;
   double fPitch, bPitch;
   double fz, mz, bz;

   double bsigma, fsigma, msigma;
   double bsigmaz, fsigmaz, msigmaz;

   init_config(config);

   /* place middle axle */
   mQ = p3dGetAxleRoll(robot, geom, x,y,theta, map, &mRoll, &mz, &mpUnk, &msigma, &msigmaz);
   if (mQ < 0.0) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: middle axle placement unstable\n");
      return -1.0;
   }

	/* 
	 * If we can't place the middle axle because we don't know what is the
	 * ground under us, it is not relevant to place other axle particulary when
	 * we take in account the position of this axle (which is not known).
	 *
	 * Just set to a default position and say that we are in unknow terrain. The
	 * cost algorithm will decide what we must do in this case
	 */
	if (mpUnk >= geom->unkMax) {
		config->x = x;
		config->y = y;
		config->theta = theta;
		config->roll = 0.0;
		config->pitch = 0.0;
		config->alpha1 = 0;
		config->alpha2 = 0;
   		config->beta3 = M_PI ;
		config->q = 1.0;
		config->pUnk = mpUnk;
		config->sigmaRoll  = 0;
   		config->sigmaPitch = 0;
   		config->sigmaz = 0;
		return config->q;
	}

   /* place front axle */
   fQ = p3dGetConstrainedAxleRoll(robot, geom, x,y,mz,theta,
				  geom->arm[0], map,
				  &fRoll, &fPitch, &fz, &fpUnk, &fsigma, &fsigmaz);
   if (fQ < 0.0) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: front axle placement unstable\n");
      return -1.0;
   }

   /* place back axle */
   bQ = p3dGetConstrainedAxleRoll(robot, geom, x,y,mz,theta,
				  -geom->arm[1], map,
				  &bRoll, &bPitch, &bz, &bpUnk, &bsigma, &bsigmaz);
   if (bQ < 0.0) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: back axle placement unstable\n");
      return -1.0;
   }

   /* quality, worst case */
   config->q = MIN(mQ, MIN(bQ, fQ));

   /* position wrt. the map center */
   config->x = x;
   config->y = y;
   config->z = mz;
   config->theta = theta;

   /* angles */
   config->pitch = (fPitch + bPitch)/2.0;
   config->roll = mRoll;
   config->alpha1 = fRoll - mRoll;
   config->alpha2 = bRoll - mRoll;
   config->beta3 = M_PI + fPitch - bPitch;
   
   /* unkown */
   config->pUnk = MAX(mpUnk, MAX(fpUnk, bpUnk));

   return config->q;
}


/*
 * p3dDubbins -----------------------------------------------------------
 *
 * Compute Dubbins distance to reach Gx, Gy from x,y,theta with radius r 
 */

static int	p3dDubbins1(double gx, double gy, double r, double y0,
			double *theta);

double
p3dDubbins(double gx, double gy, double x, double y, double theta, double r)
{
   double cost = cos(theta), sint = sin(theta);
   double Gx, Gy, tx, ty;
   double angle1, angle2, d;

   /* goal coordinates in robot frame */
   Gx =  (gx-x)*cost + (gy-y)*sint;
   Gy = -(gx-x)*sint + (gy-y)*cost;

#if 0	/* Cartesian distance */
   return sqrt(SQ(Gx)+SQ(Gy));

#else	/* Dubbins */

   if (p3dDubbins1(Gx, Gy, r, r, &angle1)) {
      if (p3dDubbins1(Gx, Gy, r, -r, &angle2)) {
	 fputs("p3d: dubbins failed\n", stderr);
	 return 1e5;
      }
      tx = r*cos(angle2);
      ty = r*sin(angle2)-r;
      d = r*fabs(M_PI_2 - angle2);

   } else {
      if (p3dDubbins1(Gx, Gy, r, -r, &angle2)) {
	 tx = r*cos(angle1);
	 ty = r*sin(angle1)+r;
	 d = r*fabs(angle1 + M_PI_2);
      } else {
	 if (fabs(M_PI_2 - angle2) > fabs(angle1 + M_PI_2)) {
	    tx = r*cos(angle1);
	    ty = r*sin(angle1)+r;
	    d = r*fabs(angle1 + M_PI_2);
	 } else {
	    tx = r*cos(angle2);
	    ty = r*sin(angle2)-r;
	    d = r*fabs(M_PI_2 - angle2);
	 }
      }
   }

   return d + sqrt(SQ(Gx-tx)+SQ(Gy-ty));
#endif
}


static int
p3dDubbins1(double gx, double gy, double r, double y0, double *theta)
{
   double delta, tant, sint, cost, theta1, theta2;
   double ps1, ps2;

   /* trivial solution */
   if (fabs(gy)<1e-5) { *theta = 0.0; return 0; }

   /* abort if goal is inside circle */
   if (SQ(gx)+SQ(gy-y0) <= SQ(r)) return -1;

   /* XXX perturbation if degenerated (solution will be approx.) */
   if (fabs(fabs(gy-y0) - r) <= 1e-5) gy-=1e-5;

   /* compute tan(theta) */
   delta = SQ(gy-y0)+SQ(gx)-SQ(r);

   /* first solution */
   tant = (-gx*(gy-y0) + r*sqrt(delta))/(SQ(gy-y0)-SQ(r));
   theta1 = atan(tant); sint = sin(theta1); cost = cos(theta1);

   /* +- M_PI ? */
   ps1 =  (gx-r*cost)*cost + (gy-r*sint-y0)*sint;
   ps2 = -(gx+r*cost)*cost - (gy+r*sint-y0)*sint;
   if (fabs(ps1) > fabs(ps2)) theta1 += M_PI;

   /* second solution */
   tant = (-gx*(gy-y0) - r*sqrt(delta))/(SQ(gy-y0)-SQ(r));
   theta2 = atan(tant); sint = sin(theta2); cost = cos(theta2);

   /* +- M_PI ? */
   ps1 =  (gx-r*cost)*cost + (gy-r*sint-y0)*sint;
   ps2 = -(gx+r*cost)*cost - (gy+r*sint-y0)*sint;
   if (fabs(ps1) > fabs(ps2)) theta2 += M_PI;

   /* get the best */
   sint = sin(theta1); cost = cos(theta1);
   ps1 = -(gx-r*cost)*sint + (gy - r*sint - y0)*cost;
   if (y0*ps1 > 0)
      *theta = theta1;
   else
      *theta = theta2;

   return 0;
}

/*
 * p3dGetConstrainedAxleRoll --------------------------------------------
 *
 * Compute the axle roll at position "x+len,y,z,theta" wrt. the center of
 * "map".
 *
 * Return the quality of placement (average of the one of the two wheels)
 * in [0;1] (1.0=best, <0.0=error). "roll" is set to the roll angle,
 * "height" to the elevation of middle point and "pitch" is pitch the
 * angle measured from the horizontal plane. pUnk is the percentage of
 * unknown terrain under the axle.  
 */

static double
p3dGetConstrainedAxleRoll(P3DRobot *robot, P3DRobotGeometry *geom, 
			  double x, double y, double z, double theta,
			  double len, P3D_MAP *map,
			  double *roll, double *pitch, double *height,
			  double *pUnk, double *var, double *var_z)
{
   double aQ;
   double xc, yc;
   double cost = cos(theta), sint = sin(theta);
   double cosp;
   int nIter = 0;
   double damping = 1.0/robot->damping;
   double iPitch, dPitch;
   double flen = fabs(len);

   /* initialy horizontal */
   iPitch = dPitch = 0.0;

   do {
      iPitch += damping*dPitch;
      cosp = cos(iPitch);
      xc = len*cost*cosp + x;
      yc = len*sint*cosp + y;

      /* place the axle */
      aQ = p3dGetAxleRoll(robot, geom,
			  xc, yc, theta, map, roll, height, pUnk, var, var_z);
      if (aQ < 0.0) return -1.0;

      /* if too much unknown don't bother iterating */
      if (*pUnk > geom->unkMax) break;

      /* pitch increment */
      if( fabs(z-*height) <= flen)
	dPitch = asin((z-*height)/flen) - iPitch;
      else /* Pitch will be >= Pi/2, forget about it */
	dPitch = M_PI_2;

      damping *= robot->damping;

   } while ( fabs(damping*dPitch) >= robot->rollStep &&
	     ++nIter < robot->nIter &&
	     fabs(dPitch) < M_PI_2 );

   /* unstable */
   if (nIter >= robot->nIter || fabs(dPitch) >= M_PI_2) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: constrained axle placement unstable\n");
      return -1.0;
   }

   *pitch = iPitch + dPitch;
   return aQ;
}


/*
 * p3dGetAxleRoll -------------------------------------------------------
 *
 * Compute the axle roll when it is placed at position "x,y", orientation
 * "theta" wrt. the center of "map".
 *
 * Return the quality of placement (average of the one of the two wheels)
 * in [0;1] (1.0=best, <0.0=error). roll is set to the roll angle, height
 * to the elevation of middle point and pUnk is the percentage of unknown
 * terrain under the axle. 
 */

static double
p3dGetAxleRoll(P3DRobot *robot, P3DRobotGeometry *geom, 
	       double x, double y, double theta, P3D_MAP *map,
	       double *roll, double *height, double *pUnk, double *var, double *var_z)
{
   double lQ, rQ;
   double lpUnk, rpUnk;
   double lWu, lWv, lWh, rWu, rWv, rWh;
   double lsigma, rsigma;
   double iRoll, dRoll;
   double damping = 1.0/robot->damping;
   int nIter = 0;
   int wI;

   /* Initial roll = 0 */
   iRoll = dRoll = 0.0;

   /* iterate */
   do {
      iRoll += damping*dRoll;
      wI = (iRoll-robot->rollMin)/robot->rollStep + 0.5;
      if (wI < 0 || wI >= robot->nProfiles) return -1.0;

      /* place the two wheels */
      lQ = p3dGetWheelHeight(robot->lWheel[wI], x, y, theta, map,
			     &lWu, &lWv, &lWh, &lpUnk, &lsigma);
      if (lQ < 0.0) return -1.0;

      rQ = p3dGetWheelHeight(robot->rWheel[wI], x, y, theta, map,
			     &rWu, &rWv, &rWh, &rpUnk, &rsigma);
      if (rQ < 0.0) return -1.0;

      /* if too much unknown don't bother iterating */
      if (lpUnk >= geom->unkMax || rpUnk >= geom->unkMax) break;

      /* roll increment */
      dRoll = atan2(lWh-rWh, lWv-rWv);
      damping *= robot->damping;

   } while (fabs(damping*dRoll) >= robot->rollStep && 
	    ++nIter < robot->nIter);

   /* unstable */
   if (nIter >= robot->nIter) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: simple axle placement unstable\n");
      return -1.0;
   }

   /* special case for new terrain */
   if (lpUnk >= geom->unkMax || rpUnk >= geom->unkMax) {
     *roll = 0.0;
#if defined(ATRV) || defined(TROOPER) || defined(RMP400) /* rad0 (wheel radius) should be the z of axle on flat terrain */
     *height = geom->rad[0];
#else
     *height = 0.0;
#endif
      *pUnk = MAX(lpUnk, rpUnk);
      return 1.0;
   }

   *roll = iRoll + dRoll;
   *height = (rWh*lWv - lWh*rWv) / (lWv - rWv);

   /** NEW: var_roll for the axle (variance = sigma^2) **/
   /* delta_z = lWh-rWh and l = lWv-rWv */
   *var = SQ(lWv-rWv)/(SQ(lWv-rWv)+SQ(lWh-rWh)) * (SQ(lsigma)+SQ(rsigma));
   /** Approximation... **/
   *var_z = (SQ(lsigma)+SQ(rsigma))/4.0;

   /* quality, worst case */
   *pUnk = MAX(lpUnk, rpUnk);
   return MIN(lQ, rQ);
}


/*
 * p3dGetWheelHeight ----------------------------------------------------
 *
 * Compute the "wheel" height when placed at position "x,y", orientation
 * "theta" wrt. the center of "map".
 *
 * Return the quality of placement in [0;1] (1.0=best, <0.0=error) and
 * wu,wv,wh is set to the 3D contact point coordinates wrt. the wheel
 * reference point. pUnk is the percentage of unknown terrain under the
 * wheel.
 */

static double
p3dGetWheelHeight(P3DElement *wheel,
		  double x, double y, double theta, P3D_MAP *map,
		  double *wu, double *wv, double *wh, double *pUnk, double *sigma)
{
#define FIX	16
#define UNDEF	((int)0x7FFFFFFF)
   double cost = cos(theta), sint = sin(theta);
   double u,v,h, hc;
   int l, c, newl, newc, i, j, isUnknown, nUnknown;
   int Bx, By, Vx, Vy, Ux, Uy, sBx, sBy; /* 16:16 fixed point */
   int Mx, My;
   float *wp;

   P3D_MAP_CELL  *mapCell;  /* current map cell */
   
   /**** NEW: To compute the sigma_z average ****/
   float sigma_z_sum=0.0, sigma_average;
   int   num_sigma_z=0;

   /* set this now */
   *pUnk = 0.0;

   /* position of top left corner of wheel (B) in map */
   u = (wheel->y*cost - wheel->x*sint + x) / map->scale;
   v = (wheel->y*sint + wheel->x*cost + y) / map->scale;
   l = floor(u);  /* "Abscisse pixel" on map (No line) */
   if (l < 0 || l >= map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);  /* "Ordonnee pixel" on map (No col.) */
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* initialize pointers on dem */
   mapCell = &(map->cells[l][c]);

   /* bitmap coords of B (fixed point) */
   Bx = u * (1<<FIX);
   By = v * (1<<FIX);

   /* position of bottom left corner of wheel (V) in map */
   u = ((wheel->y - wheel->h*wheel->scale)*cost - wheel->x*sint + x)
      / map->scale;
   v = ((wheel->y - wheel->h*wheel->scale)*sint + wheel->x*cost + y)
      / map->scale;
   l = floor(u);
   if (l < 0 || l >= map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* increment from B to V (bitmap cells, fixed point) */
   Vx = (u * (1<<FIX) - Bx) / wheel->h;
   Vy = (v * (1<<FIX) - By) / wheel->h;

   /* position of top right corner of wheel (U) */
   u = (wheel->y*cost - (wheel->x - wheel->w*wheel->scale)*sint + x)
      / map->scale;
   v = (wheel->y*sint + (wheel->x - wheel->w*wheel->scale)*cost + y)
      / map->scale;
   l = floor(u);
   if (l < 0 || l > map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* increment from B to U, (bitmap cells, fixed point) */
   Ux = (u * (1<<FIX) - Bx) / wheel->w;
   Uy = (v * (1<<FIX) - By) / wheel->w;

   /* l and c hold the current line and column inside dem */
   l = Bx>>FIX;		c = By>>FIX;

   /* top left corner is inside dem */
   isUnknown = 0;	nUnknown = 0;

   /* loop on wheel pixels and compute maximum elevation */
   Mx = My = UNDEF;
   h = -1e99;
   wp = wheel->height;
   for (i=0; i<wheel->h; i++) {			/* lines */

      sBx = Bx;	sBy = By;

      for(j=0; j<wheel->w; j++, wp++) {		/* columns */

	 if (*wp != P3D_UNDEF_ELEM) {
	    if (!isUnknown && mapCell->state != P3D_MAP_CELL_EMPTY) {
	       hc = mapCell->z - *wp;
	       if (hc > h) {
		  h = hc;
		  Mx = j;
		  My = i;
	       }
	    } else nUnknown++;
	 }

	 /* new pixel */
	 Bx += Ux; newl = Bx>>FIX;
	 By += Uy; newc = By>>FIX;

	 /* line changed? */
	 if (newl != l) {
	   mapCell = &(map->cells[newl][c]);
	    l = newl;
	    isUnknown = (c < 0 || c >= map->nbCols);
	 }
	 /* column changed? */
	 if (newc != c) {
	   mapCell = &(map->cells[l][newc]);
	    c = newc;
	    isUnknown = (c < 0 || c >= map->nbCols);
	 }
	 /**** NEW: get the sigma_z under the wheel to compute the average ****/
	 /* the number of sigma considered is counted wrt number of Wheel pixels */
	 if (*wp != P3D_UNDEF_ELEM)
	   {
	     sigma_z_sum += mapCell->sigma_z;
	     num_sigma_z++;
	   }
	 /************************************************************/
      }

      /* new pixel */
      Bx = sBx + Vx;	newl = Bx>>FIX;
      By = sBy + Vy;	newc = By>>FIX;

      /* line changed? */
      if (newl != l) {
	mapCell = &(map->cells[newl][c]);
	l = newl;
	isUnknown = (c < 0 || c >= map->nbCols);
      }
      /* column changed? */
      if (newc != c) {
	mapCell = &(map->cells[l][newc]);
	c = newc;
	isUnknown = (c < 0 || c >= map->nbCols);
      }
   }


   /* position of max height */
   if (Mx == UNDEF || My == UNDEF) {
      *pUnk = 1.0;
      return 1.0;
   }

   *wu = wheel->y - My*wheel->scale;
   *wv = wheel->x - Mx*wheel->scale;
   *wh = h;
   *pUnk = (double)nUnknown/wheel->nElem;

   /**** NEW: compute the average of sigma_z under the wheel ****/
   sigma_average = sigma_z_sum / ( (float) num_sigma_z);
   *sigma = (double) sigma_average;

   /* placement quality (not used at the moment) */
   return 1.0;
}


/*
 * p3dWarnUnstable ------------------------------------------------------
 *
 * For debugging purposes.
 */

static void
p3dWarnUnstable(double x, double y, double theta, const char *message)
{
#if VERBOSE  /* printf only if Verbose mode asked */
  fprintf(stderr, "p3d: %s on %f %f %f\n", message, x, y, theta);
#else/* VERBOSE */
  (void)x;
  (void)y;
  (void)theta;
  (void)message;
#endif
}


/*
 * p3dGetArcPos ---------------------------------------------------------
 *
 * Compute robot horizontal position and orientation given an arc
 * (r=curvature), a distance to follow (ds) and an initial position (x0,
 * y0, theta0). 
 */

void
p3dGetArcPos(double r, double ds, double x0, double y0, double theta0,
	     double *x, double *y, double *theta)
{
   double dx, dy, dth;
   double cth, sth;
   
   dth = r*ds;
   
   if (fabs(r)<1e-5) {
      dx = ds;
      dy = 0.0;
   } else {
      dx = sin(dth)/r;
      dy = (1-cos(dth))/r;
   }

   cth = cos(theta0); sth = sin(theta0);

   *x = cth*dx - sth*dy + x0;
   *y = sth*dx + cth*dy + y0;
   *theta = dth + theta0;
}



/**********************************************************************/


/*
 * p3dPlaceRobotDala --------------------------------------------------------
 *
 * Compute robot configuration for position "x,y", orientation "theta"
 * wrt. the center of "map".
 *
 * Return the quality of placement (minimum of the one of the three
 * axles). "roll" and "pitch" are the configuration of the middle axle,
 * "alpha1" the *relative* roll between the front and back axle
 * "alpha2" is not used
 * and "beta" is not used
 * pUnk is the max percentage of unknown terrain for each wheel.
 */

double
p3dPlaceRobotDala(P3DRobot *robot, P3DRobotGeometry *geom,
		  double x, double y, double theta, P3D_MAP *map,
		  P3DRobotConfig *config)
{
   double fQ, bQ;
   double fpUnk, bpUnk;
   double fRoll, bRoll;
   double fPitch;
   double fz, mz, bz;
   double bx,by;
   double bvar, fvar;
   double bvarz, fvarz;
   double vard;   /* variance due to discretization (= SQ(sigmad) ) */

   double Length; /* front to back axle length */

   init_config(config);

   /* place back axle */
   /* geom->arm[1] doit etre la distance entre le centre du robot et le back axle */
   /* First compute coords of center of back axle */
   bx = -geom->arm[1]*cos(theta) + x;
   by = -geom->arm[1]*sin(theta) + y;
   bQ = p3dGetAxleRoll(robot, geom, bx,by,theta, map, &bRoll, &bz, &bpUnk, &bvar, &bvarz);
   if (bQ < 0.0) {
     p3dWarnUnstable(x, y, theta,
		     "p3d: back axle placement unstable\n");
     return -1.0;
   }

   /*
	 * If we can't place the middle axle because we don't know what is the
	 * ground under us, it is not relevant to place other axle particulary when
	 * we take in account the position of this axle (which is not known).
	 *
	 * Just set to a default position and say that we are in unknow terrain. The
	 * cost algorithm will decide what we must do in this case
	 * */
	if (bpUnk >= geom->unkMax) {
		config->x = x;
		config->y = y;
		config->theta = theta;
		config->roll = 0.0;
		config->pitch = 0.0;
		config->alpha1 = 0;
		config->alpha2 = 0;
   		config->beta3 = M_PI ;
		config->q = 1.0;
		config->pUnk = bpUnk;
		config->sigmaRoll  = 0;
   		config->sigmaPitch = 0;
   		config->sigmaz = 0;
		return config->q;
	}

   /* We have now bRoll and bz */
   
   /* place front axle */
   /* geom->arm[0] doit etre la distance entre le centre du robot et le front axle */
   /* le point de reference utilise ici sera bx,by,bz plutot, avec len=long.interaxles */
   fQ = p3dGetConstrainedAxleRoll(robot, geom, bx,by,bz,theta,
				  geom->arm[0]+geom->arm[1], map,
				  &fRoll, &fPitch, &fz, &fpUnk, &fvar, &fvarz);
   if (fQ < 0.0 || isnan(fPitch)) {
      p3dWarnUnstable(x, y, theta,
		      "p3d: front axle placement unstable\n");
#if TEMP
      if(isnan(fPitch))
	fprintf(stderr, "fPitch is nan\n");
#endif
      return -1.0;
   }


   /* fPitch is Pitch, so we can calculate mz */
/*    mz = bz + geom->arm[1]*sin(fPitch); */
   /* The Robot's center is just in the middle between the back and forward axles */
   mz = (bz+fz)/2.0;

   /* quality, worst case */
   config->q = MIN(bQ, fQ);

   /* position wrt. the map center */
   /*--- Plutot wrt. Robot's current position (?)--*/
   config->x = x;
   config->y = y;
   config->z = mz;
   config->theta = theta;

   /* angles */
   config->pitch = fPitch ;
   /* if the current placement is accepted, roll ~= fRoll ~= bRoll */
   config->roll = (fRoll+bRoll)/2.0;  /* (ou Max(fRoll,bRoll) ? ) */
   /*** TEST Th *****************************/
/*    if( fabs(fRoll) >= fabs(bRoll) ) */
/*      config->roll = fRoll; */
/*    else */
/*      config->roll = bRoll; */
   /*** END TEST Th *************************/

   /* we should have alpha1=alpha2, considering Dala's structure */
   config->alpha1 = fRoll;
   config->alpha2 = bRoll;

   /* unkown */
   config->pUnk = MAX(fpUnk, bpUnk);

#if DEBUG
   fprintf(stderr, "config->z = %lf\n", config->z);
#endif

#if BODYTEST
   if (p3dCheckBodyCollision(robot, geom, map, config) < 0)
	   return -1.0;
#endif  /* BODYTEST */

   /* beta3 should be set to some useless constant value or simply removed from the structure */
   config->beta3 = M_PI ;
   
   /*** NEW: computation of a sigma for that placement ***/
   /** variance due to discretization (cf rollStep) must be added **/
   vard = SQ( robot->rollStep*P3D_DISCRETISATION_FACTOR );
   bvar = SQ(bvar) + vard;
   fvar = SQ(fvar) + vard;
   /** NB: should be optimized: no need to recompute that factor everytime... **/

   config->sigmaRoll  = sqrt( (bvar+fvar)/4 );
   Length = geom->arm[0] + geom->arm[1];
   config->sigmaPitch = sqrt( (bvar+fvar) / (SQ(Length) - SQ(bz-fz)) );
   config->sigmaz = sqrt( (bvarz+fvarz)/4 );


   return config->q;
}







/*
 * p3dDalaBodyCollisionTest ----------------------------------------------------
 *
 * Check if le fond de la caisse de Dala entre en collision avec le DEM,
 * comparing its heights with the DEM heights when the Robot is placed at position "x,y",
 * orientation "theta" wrt. the center of "map".
 *
 * Return the quality of placement in [0;1] (1.0=best, <0.0=error),
 * wz the minimum of z distance between caisse and dem (if < 0, collision) and
 * wu,wv the coordinates of this minimum wrt. the "caisse" reference point.
 * pUnk is the percentage of unknown terrain under the "fond de caisse".
 *
 * Input: x,y,z are the coordinate of the robot's center
 */

#if ((defined(ATRV) || defined(TROOPER) || defined(RMP400)) && BODYTEST)
static double
p3dBodyCollisionTest(P3DElement *body,
			 double x, double y, double z, double theta, P3D_MAP *map,
			 double *wu, double *wv, double *wh, double *pUnk)
{
#define FIX	16
#define UNDEF	((int)0x7FFFFFFF)
#define P3D_MAX_BODY_HEIGHT  (10000.0)
#define P3D_MIN_BODY_HEIGHT (0.0)  /* Min de z(body)-z(dem) tolere */
  /* TODO: Mettre ca dans la structure P3DRobotGeometry ? */
   double cost = cos(theta), sint = sin(theta);
   double u,v,h;
   int l, c, newl, newc, i, j, isUnknown, nUnknown;
   int Bx, By, Vx, Vy, Ux, Uy, sBx, sBy; /* 16:16 fixed point */
   int Mx, My;
   float *wp;
   /*-----*/
   P3D_MAP_CELL  *mapCell;  /* current map cell */

   double delta_z, minDelta_z=P3D_MAX_BODY_HEIGHT;


   /* If body is empty, exit and return a 0 to indicate that */
   if (body == NULL) { errno = EINVAL; return 0.0; }

   /* set this now */
   *pUnk = 0.0;

   /* position of back left corner of body (B) */
   u = (body->y*cost - body->x*sint + x) / map->scale;
   v = (body->y*sint + body->x*cost + y) / map->scale;
   
   l = floor(u);
   if (l < 0 || l >= map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* initialize pointer on map */
   mapCell = &(map->cells[l][c]);

   /* bitmap coords of B (fixed point) */
   Bx = u * (1<<FIX);
   By = v * (1<<FIX);

   /* position of front left corner of body (V) */
   u = ((body->y - body->h*body->scale)*cost - body->x*sint + x)
      / map->scale;
   v = ((body->y - body->h*body->scale)*sint + body->x*cost + y)
      / map->scale;
   l = floor(u);
   if (l < 0 || l >= map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* increment from B to V (bitmap cells, fixed point) */
   Vx = (u * (1<<FIX) - Bx) / body->h;
   Vy = (v * (1<<FIX) - By) / body->h;

   /* position of back right corner of body (U) */
   u = (body->y*cost - (body->x - body->w*body->scale)*sint + x)
      / map->scale;
   v = (body->y*sint + (body->x - body->w*body->scale)*cost + y)
      / map->scale;
   l = floor(u);
   if (l < 0 || l > map->nbLines) { errno = EINVAL; return -1; }
   c = floor(v);
   if (c < 0 || c > map->nbCols) { errno = EINVAL; return -1; }

   /* increment from B to U, (bitmap cells, fixed point) */
   Ux = (u * (1<<FIX) - Bx) / body->w;
   Uy = (v * (1<<FIX) - By) / body->w;

   /* l and c hold the current line and column inside dem */
   l = Bx>>FIX;		c = By>>FIX;

   /* top left corner is inside dem */
   isUnknown = 0;	nUnknown = 0;


   /* loop on body pixels and compute maximum elevation */
   Mx = My = UNDEF;
   h = -1e99;
   wp = body->height;  /* height of body bottom wrt robot center (which is at z) */
   for (i=0; i<body->h; i++) {			/* lines */

      sBx = Bx;	sBy = By;

      for(j=0; j<body->w; j++, wp++) {		/* columns */

	if (*wp != P3D_UNDEF_ELEM) {
	  if (!isUnknown && mapCell->state != P3D_MAP_CELL_EMPTY) {
	    /* Get the body bottom z at position of current pixel -> z+zbody */
	    /* zbody is the z(robot centre -> body bottom) */
	    /* And compare with the corresponding z(map) */
	    delta_z = (z + *wp) - mapCell->z;

	    /* if delta_z too small -> risk of collision (=> stop) */
	    if (delta_z > P3D_MIN_BODY_HEIGHT) {
	      /* Calculation of min(delta_z) above body bottom */
	      if(delta_z < minDelta_z)
		{ /* update minDelta_z */
		  minDelta_z = delta_z;
		  Mx = j;
		  My = i;
		}
	    }
	    else 
	      {
		/* Get out of the loop -> placement too dangerous */
/* 		fprintf(stderr, "p3d: position %f %f %f: robot body collision with the ground\n", x, y, theta); */
		*wu = body->y - My*body->scale;
		*wv = body->x - Mx*body->scale;
		return -1.0;
	      }

	  } else nUnknown++;
	}

	 /* new pixel */
	 Bx += Ux; newl = Bx>>FIX;
	 By += Uy; newc = By>>FIX;

	 /* line changed? */
	 if (newl != l) {
	   mapCell = &(map->cells[newl][c]);
	   l = newl;
	   isUnknown = (c < 0 || c >= map->nbCols);
	 }
	 /* column changed? */
	 if (newc != c) {
	   mapCell = &(map->cells[l][newc]);
	   c = newc;
	   isUnknown = (c < 0 || c >= map->nbCols);
	 }
      }

      /* new pixel */
      Bx = sBx + Vx;	newl = Bx>>FIX;
      By = sBy + Vy;	newc = By>>FIX;

      /* line changed? */
      if (newl != l) {
	mapCell = &(map->cells[newl][c]);
	l = newl;
	isUnknown = (c < 0 || c >= map->nbCols);
      }
      /* column changed? */
      if (newc != c) {
	mapCell = &(map->cells[l][newc]);
	c = newc;
	isUnknown = (c < 0 || c >= map->nbCols);
      }
   }
   

   /* position of max height */
   if (Mx == UNDEF || My == UNDEF) {
      *pUnk = 1.0;
      return 1.0;
   }

   *wu = body->y - My*body->scale;
   *wv = body->x - Mx*body->scale;
   *wh = minDelta_z;
   *pUnk = (double)nUnknown/body->nElem;

   /* placement quality (not used at the moment) */
   return 1.0;
}

/*
 * Check Body Collision with ground
 * In input, we need to have a config with at least set x,y,z,theta
 * It is not a problem because it is the last operation
 */
static int
p3dCheckBodyCollision(P3DRobot *robot, P3DRobotGeometry *geom, 
		P3D_MAP *map, P3DRobotConfig *config)
{
	double resultCollision=0.0; /* 0.0 means no body collision test has been made */
	double body_u, body_v;    /* Coordinates of body bottom point the closest to the terrain */
	double body_dz = 0.0;    /* delta_z (body-terrain) at that point */
	double body_pUnk;  /* Unknown percentage under body */
	int  numRoll, numPitch;  /* indexes of for body */

	/*-------- Collision test of robot's body with dem map--------*/
	/* Compute the indexes of Body bitmap */
	/* (closest integers to actual double roll & pitch) */
	numRoll = (int) floor( (config->roll - robot->bodyRollMin)/robot->bodyRollStep +0.5 );
	numPitch = (int) floor( (config->pitch - robot->bodyPitchMin)/robot->bodyPitchStep +0.5 );

	/* If roll or pitch are over limits, do not try to check body collision */ 
	/* (it is useless and there are no body bitmaps prepared for ) */
	if(numRoll >= robot->nBodyRoll || numPitch >= robot->nBodyPitch
			|| numRoll < 0 || numPitch < 0 )
	{
		fprintf(stderr, 
				"p3dCheckBodyCollision: WARNING: numRoll or numPitch over limits\n");
		fprintf(stderr, 
				"p3dCheckBodyCollision: numRoll=%d, numPitch=%d\n", numRoll, numPitch);
		fprintf(stderr, 
				"p3dCheckBodyCollision: config->roll=%lf, config->pitch=%lf\n",
				config->roll, config->pitch);
		fprintf(stderr, "p3dCheckBodyCollision: Could not compute delta_z\n");
		body_dz = 0.0;
	}
	else
	{
		/* If unknown terrain, should not test body collision with ground */
		/* besides, it would not work as zAxle(s) have been set to zero */
		if (config->pUnk < geom->unkMax) {
			resultCollision = 
				p3dBodyCollisionTest(robot->body[numRoll][numPitch], 
						config->x, config->y, config->z, config->theta, map,
						&body_u, &body_v, &body_dz, &body_pUnk);
#if DEBUG
			fprintf(stderr, "p3dCheckBodyCollision: delta_z = %lf\n", body_dz);
#endif
		}
		else {   /* mostly unknown terrain */
			body_dz = geom->groundClear;  /* Default: dz = robot ground clearance */
			body_pUnk = 1.0;
		}
	}


	if(resultCollision < 0.0) {
		p3dWarnUnstable(config->x, config->y, config->theta,
				"p3d: Body collision with ground\n");
		return -1;
	}

	/* Record evaluation of min. height between robot body bottom and ground */
	config->dzBodyGround = body_dz;

	return 0;
}
#endif /* ATRV && BODYTEST */

double  p3dPlaceRobotTrooper(P3DRobot *robot, P3DRobotGeometry *geom,
							  double x, double y, double theta,
							  P3D_MAP *map, P3DRobotConfig *config)
{
	double res;
	res = p3dPlaceRobot(robot, geom, x, y, theta, map, config);
	if (res < 0.0) 
		return res;

#if BODYTEST
	if ( p3dCheckBodyCollision(robot, geom, map, config) < 0)
		return -1.0;
#endif

	return config->q;
}




